
import os
from pickletools import read_uint1
import re
import sys
import imp
from functools import reduce
from urllib import response

from jinja2.filters import K
from .constants import *
from .logger import Logger
from .dict_utils import *
from pathlib import Path
from .file_utils import *
from .exceptions import DSOException
from .stages import Stages
from .contexts import Context, ContextScope, ContextMode
from .enum_utils import OrderedEnum
from .remote_configs import RemoteConfig

key_regex_pattern = {
    "get": r"^[a-zA-Z]([.a-zA-Z0-9_-]*[a-zA-Z0-9])?$",
    "add": r"^[a-zA-Z]([.a-zA-Z0-9_-]*[a-zA-Z0-9*])?$"
}


class ConfigOrigin(OrderedEnum):
    All = 10
    Local = 20
    Remote = 30


_init_config = {
    'kind': 'dso/application',
    'version': 1,
    'namespace': 'generic',
    'application': '',
    'config': {
        'provider': {
            'id': 'local/v1',
            'spec': {}
        },
    },
    'parameter': {
        'provider': {
            'id': 'local/v1',
            'spec': {}

        },
    },
    'secret': {
        'provider': {
            'id': 'shell/v1',
            'spec': {}

        },
    },
    'template': {
        'provider': {
            'id': 'local/v1',
            'spec': {}

        },
    },    
}

_default_config = {
    'kind': 'dso/application',
    'version': 1,
    'namespace': 'generic',
    'application': '',
    'config': {
        'provider': {
            'id': '',
            'spec': {}
        },
    },
    'parameter': {
        'provider': {
            'id': '',
            'spec': {}
        },
    },
    'secret': {
        'provider': {
            'id': '',
            'spec': {}
        },
    },
    'template': {
        'provider': {
            'id': '',
            'spec': {}
        },
        'renderPath': {}
    },
    'artifactory': {
        'provider': {
            'id': '',
            'spec': {}
        },
    },
    'package': {
        'provider': {
            'id': '',
            'spec': {},
            'essential' : [],
            'optional': []
        },
    },
    'release': {
        'provider': {
            'id': '',
            'spec': {},
            'essential' : [],
            'optional': []
        },
    },
    'network':{
        'subnetPlanGroup': '',
        'subnetPlan': '',
        'selector': {}
    }, 

}

all_services = ['config', 'parameter', 'secret', 'template', 'artifactory', 'package', 'release']


def get_init_config():
    return _init_config.copy()


def get_default_config():
    return _default_config.copy()


class ConfigService:
    @property
    def config_dir(self):
        return '.dso'

    @property
    def config_file(self):
        return 'dso.yml'

    @property
    def install_path(self):
        return os.path.dirname(os.path.abspath(__file__))

    context = Context() 
    original_context = Context()
    working_dir = ''
    local_config = {}
    local_config_rendered = {}
    local_config_file_path = ''
    local_config_dir_path = ''
    root_config = {}
    root_config_rendered = {}
    root_config_file_path = ''
    root_config_dir_path = ''
    overriden_config = {}
    merged_config = {}
    remote_config = {}
    remote_config_rendered = {}


    def validate_key(self, key, action):
        Logger.info(f"Validating key '{key}'...")
        if not key:
            raise DSOException(MESSAGES['KeyNull'])
        if key == 'dso' or key.startswith('dso.'):
            raise DSOException(MESSAGES['DSOReserverdKey'].format(key))
        if not re.match(key_regex_pattern[action], key):
            raise DSOException(MESSAGES['InvalidKeyPattern'].format(key, key_regex_pattern[action]))
        if '..' in key:
            raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '..'))


    def load(self, working_dir, config_overrides_string='', stage=None, scope=ContextScope.App, ignore_errors=False):

        if stage is None: stage = 'default'
        stage = Stages.normalize(stage)

        self.working_dir = working_dir

        ### start off with given stage, scope, and overriden config to set meta_data for subsequent rendering
        self.context = Context('default', 'default', stage, scope)
        self.apply_config_overrides(config_overrides_string)

        ### load root and local configs only without rendering
        self.load_root_config(render=False)
        self.load_local_config(render=False)
        
        if not ignore_errors:
            if not self.merged_config['namespace']:
                raise DSOException("Namespace has not been set. Run 'dso config set namespace <name>' to do so.")
            if self.merged_config['namespace'] == 'default':
                raise DSOException("Namespace name cannot be 'default'.")
            if scope == ContextScope.Namespace and not Stages.is_default_env(stage):
                raise DSOException("Numbered environments are not allowd when using the namespace scope (for simplicity). Remove the number from the given satge and try again.")
            if scope == ContextScope.Global and not Stages.is_default_env(stage):
                raise DSOException("Numbered environments are not allowd when using the root scope (for simplicty). Remove the number from the given satge and try again.")

            if not self.merged_config['application']:
                raise DSOException("Application name has not been set. Run 'dso config set application <name>' to do so.")
            if self.merged_config['application'] == 'default':
                raise DSOException("Application name cannot be 'default'.")

            self.check_version()


    @property
    def meta_vars(self):
        return {'dso': self.merged_config}

    def apply_config_overrides(self, config_overrides):
        self.overriden_config = {}
        if config_overrides:
            if isinstance(config_overrides, str):
                config_overrides = self.config_string_to_dict(config_overrides)
            configs = flatten_dict(config_overrides)           
            not_alllowed_configs = ['stage', 'scope']  ### read-only configs
            check_default_configs = ['namespace', 'application']
            for key, value in configs.items():
                if key in not_alllowed_configs:
                    Logger.warn(f"Ignored overriding configuration '{key}', the following DSO configuration settings cannot be overriden: {not_alllowed_configs}")
                if key in check_default_configs: value = value or 'default'
                Logger.debug(f"DSO configuration '{key}' was overriden to '{value}'.")
                set_item(self.overriden_config, key.split('.'), value)

        self.update_merged_config()

    def load_root_config(self, render=True):
        self.root_config = {}
        self.root_config_rendered = {}
        self.root_config_dir_path = os.path.join(Path.home(), self.config_dir)
        self.root_config_file_path = os.path.join(self.root_config_dir_path, self.config_file)

        if not os.path.exists(self.root_config_file_path):
            Logger.debug("Root configuration not found.")
            return

        Logger.debug(f"Root configuration found: path={self.root_config_file_path}")
        try:
            self.root_config = load_file(self.root_config_file_path)
        except DSOException:
            raise
        except:
            raise DSOException(MESSAGES['InvalidDSOConfigurationFile'].format(self.root_config_file_path))

        if render:
            self.render_root_config()
        else:
            self.root_config_rendered = self.root_config.copy()

        self.update_merged_config()


    def load_local_config(self, silent_warnings=False, render=False):
        self.local_config = {}
        self.local_config_rendered = {}
        self.local_config_dir_path = os.path.join(self.working_dir, self.config_dir)
        self.local_config_file_path = os.path.join(self.local_config_dir_path, self.config_file)

        if not os.path.exists(self.local_config_file_path):
            if not silent_warnings: 
                Logger.warn(MESSAGES['NoDSOConfigFound'])
            return

        Logger.debug(f"Local configuration found: path={self.local_config_file_path}")

        try:
            self.local_config = load_file(self.local_config_file_path)
        except DSOException:
            raise
        except:
            raise DSOException(MESSAGES['InvalidDSOConfigurationFile'].format(self.local_config_file_path))

        if render:
            self.render_local_config()
        else:
            self.local_config_rendered = self.local_config.copy()

        self.update_merged_config()


    def load_remote_config(self, uninherited=False, filter=None):
        self.remote_config = {}
        self.remote_config = deflatten_dict({x['Key']: x['Value'] for x in RemoteConfig.list(uninherited=uninherited, filter=filter)})
        self.update_merged_config()


    def render_root_config(self, silent=False):
        ### call update using not rendered config for in-place rendering of the root config iteself
        self.update_merged_config(rendered=False)
        ### now do the rendering
        # self.root_config_rendered = load_file(self.root_config_file_path, pre_render_values=self.merged_config)
        self.root_config_rendered = render_dict_values(self.root_config, values=self.merged_config, silent=silent)
        # self.update_merged_config()


    def render_local_config(self, silent=False):
        ### call update using not rendered config for in-place rendering of the root config iteself
        self.update_merged_config(rendered=False)
        ### now do the rendering
        # self.local_config_rendered = load_file(self.local_config_rendered, pre_render_values=self.merged_config)
        self.local_config_rendered = render_dict_values(self.local_config, values=self.merged_config, silent=silent)
        # self.update_merged_config()


    def render_remote_config(self, silent=False):
        ### call update using not rendered config for in-place rendering of the remote config iteself
        self.update_merged_config(rendered=False)
        ### now do the rendering
        # self.local_config_rendered = load_file(self.local_config_rendered, pre_render_values=self.merged_config)
        self.remote_config_rendered = render_dict_values(self.remote_config, values=self.merged_config, silent=silent)
        # self.update_merged_config()


    def dict_to_config_string(self, dic):
        flat = flatten_dict(dic)
        return reduce(lambda x, y: f"{x[0]}={x[1]},{y[0]}={y[1]}", flat)


    def config_string_to_dict(self, config_string):
        if not config_string: return {}

        result = {}
        try:
            configs = config_string.split(',')
            for config in configs:
                if not config: continue
                key = config.split('=')[0].strip()
                value = config.split('=')[1].strip()
                set_item(result, key.split('.'), value)
        except:
            raise DSOException(MESSAGES['InvalidDSOConfigOverrides'].format(config_string))

        return result

    def update_merged_config(self, use_defaults=True, use_calculated=True, use_root=True, use_local=True, use_remote=True, rendered=False):
        self.merged_config = {}
        if use_defaults:
            self.merged_config = get_default_config()

        if use_remote:
            if rendered:
                self.render_remote_config()                
                merge_dicts(self.remote_config_rendered, self.merged_config)
            else:
                merge_dicts(self.remote_config, self.merged_config)                
        if use_root:
            if rendered:
                self.render_root_config()
                merge_dicts(self.root_config_rendered, self.merged_config)
            else:
                merge_dicts(self.root_config, self.merged_config)                
        if use_local:
            if rendered:
                self.render_local_config()
                merge_dicts(self.local_config_rendered, self.merged_config)
            else:
                merge_dicts(self.local_config, self.merged_config)                
        
        if use_defaults:
            ### save provider ids set in app config before merging with service providers' default config
            saved_service_providers = {}
            for service in all_services:
                pid = self.get_provider_id(service)
                if pid:
                    saved_service_providers[service] = {
                        'id' : pid,
                        'spec' : self.get_provider_spec(service)
                    }

            ### add missing default specs for each service
            for service in all_services:
                pid = self.get_provider_id(service)
                if pid:
                    ### merge saved spec with the default if only same provider
                    if saved_service_providers[service]['id'] == pid:
                        self.merged_config[service]['provider']['spec'] = self.get_provider_default_spec(service, pid)
                        merge_dicts(saved_service_providers[service]['spec'], self.merged_config[service]['provider']['spec'])
                    else:
                        self.merged_config[service]['provider']['spec'] = self.get_provider_default_spec(service, pid)

        ### always apply provided overriding configs
        merge_dicts(self.overriden_config, self.merged_config)

        ### update context
        self.context = Context(self.merged_config.get('namespace', ''), self.merged_config.get('application', ''), self.stage, self.scope)

        ### add calculated configs on the fly at last
        if use_calculated:
            self.merged_config['context'] = {
                'effective': {
                    'namespace': self.namespace,
                    'application': self.application,
                    'stage': self.short_stage,
                    'env': self.env,
                    'scope': str(self.scope)
                },
                'target': {
                    'namespace': self.get_namespace(ContextMode.Target),
                    'application': self.get_application(ContextMode.Target),
                    'stage': self.get_stage(ContextMode.Target, short=True),
                    'env': Stages.parse_env(self.get_stage(ContextMode.Target)),
                    'scope': str(self.scope)
                },
            }
            self.merged_config['stage'] = self.get_stage(ContextMode.Target, short=True)
            self.merged_config['env'] = Stages.parse_env(self.get_stage(ContextMode.Target))
            self.merged_config['scope'] = str(self.scope)


    def check_version(self):
        if not int(self.merged_config['version']) == int(_default_config['version']):
            if int(self.merged_config['version']) > int(_default_config['version']):
                Logger.warn(MESSAGES['DSOConfigNewer'].format(_default_config['version'], self.merged_config['version']))
            else:
                Logger.warn(MESSAGES['DSOConfigOlder'].format(_default_config['version'], self.merged_config['version']))


    def get_provider_default_spec(self, provider, provider_id):
        if f'{provider}/{provider_id}' in sys.modules:
            provider = sys.modules[f'{provider}/{provider_id}']
        else:
            providerPackagePath = os.path.join(self.install_path, 'provider', f'{provider}/{provider_id}')
            if not os.path.exists(providerPackagePath):
                raise DSOException(f"Provider '{provider}/{provider_id}' not found.")
            provider = imp.load_package(f'{provider}/{provider_id}', providerPackagePath)
        
        return provider.get_default_spec()


    def save_local_config(self):
        os.makedirs(self.local_config_dir_path, exist_ok=True)
        save_data(self.local_config, self.local_config_file_path)

    def save_root_config(self):
        os.makedirs(self.root_config_dir_path, exist_ok=True)
        save_data(self.root_config, self.root_config_file_path)


    @property
    def namespace(self):
        return self.get_namespace()

    def get_namespace(self, source=ContextMode.Effective):
        if source == ContextMode.Target:
            result = self.context.target[0]
        elif source == ContextMode.Effective:
            result = self.context.effective[0]
        
        return result


    @property
    def application(self):
        return self.get_application()


    def get_application(self, source=ContextMode.Effective):
        if source == ContextMode.Target:
            result = self.context.target[1]
        elif source == ContextMode.Effective:
            result = self.context.effective[1]
        
        return result


    @property
    def stage(self):
        return self.get_stage()

    def get_stage(self, source=ContextMode.Effective, short=False):
        if source == ContextMode.Target:
            result = self.context.target[2]
        elif source == ContextMode.Effective:
            result = self.context.effective[2]
        
        if short:
            result = Stages.shorten(result)
        else:
            result = Stages.normalize(result)
        
        return result

    @property
    def short_stage(self):
        return self.get_stage(short=True)

    @property
    def env(self):
        return Stages.parse_env(self.get_stage())


    @property
    def scope(self):
        return self.context.scope


    def get_provider_id(self, service):
        try:
            result = self.merged_config[service]['provider']['id'] or os.getenv(f"DSO_{service.upper()}_PROVIDER")
        except:
            raise DSOException(f"Invalid dso configuration schema: {service}")
        return result

    @property
    def config_provider(self):
        return self.get_provider_id('config')

    @property
    def parameter_provider(self):
        return self.get_provider_id('parameter')

    @property
    def secret_provider(self):
        return self.get_provider_id('secret')

    @property
    def template_provider(self):
        return self.get_provider_id('template')

    @property
    def artifactory_provider(self):
        return self.get_provider_id('artifactory')

    @property
    def package_provider(self):
        return self.get_provider_id('package')


    @property
    def release_provider(self):
        return self.get_provider_id('release')


    def get_provider_spec(self, service, key=None):
        try:
            result = self.merged_config[service]['provider']['spec']
        except KeyError:
            raise DSOException("Invalid dso configuration schema.")
        if not key:
            return result
        elif key in result:
            return result[key]
        else:
            return ''

    def config_spec(self, key=None):
        return self.get_provider_spec('config', key)

    def parameter_spec(self, key=None):
        return self.get_provider_spec('parameter', key)

    def secret_spec(self, key=None):
        return self.get_provider_spec('secret', key)

    def template_spec(self, key):
        return self.get_provider_spec('template', key)

    def artifactory_spec(self, key):
        return self.get_provider_spec('artifactory', key)

    def package_spec(self, key):
        return self.get_provider_spec('package', key)

    def network(self, key=None):
        try:
            result = self.merged_config['network']
        except KeyError:
            raise DSOException("Invalid dso configuration schema.")
        if not key:
            return result
        elif key in result:
            return result[key]
        else:
            return ''

    def get_template_render_paths(self, key=None):
        result = get_item(self.local_config, ['template', 'renderPath'])[0] or {}
        if not key:
            return result
        else:
            return {x:result[x] for x in result if x==key}

    def register_template_custom_render_path(self, key, render_path):
        result = get_item(self.local_config, ['template', 'renderPath'])[0] or {}
        # if os.path.isabs(render_path):
        #     raise DSOException(MESSAGES['AbsTemplateRenderPath'].format(render_path))
        # if os.path.isdir(render_path):
        #     raise DSOException(MESSAGES['InvalidRenderPathExistingDir'].format(render_path))
        result[key] = render_path
        self.local_config['template']['renderPath'] = result
        self.save_local_config()

    def unregister_template_custom_render_path(self, key):
        result = get_item(self.local_config, ['template', 'renderPath'])[0]
        if key in result:
            self.local_config['template']['renderPath'].pop(key)
            self.save_local_config()


    def init(self, working_dir, custom_init_config, config_overrides_string, override_inherited=False):
        Logger.info("Initializing DSO configurations...")
        self.working_dir = working_dir
        self.local_config_dir_path = os.path.join(self.working_dir, self.config_dir)
        self.local_config_file_path = os.path.join(self.local_config_dir_path, self.config_file)
        self.root_config_dir_path = os.path.join(Path.home(), self.config_dir)
        self.root_config_file_path = os.path.join(self.root_config_dir_path, self.config_file)

        # if config_scope == ConfigScope.Local:
        if os.path.exists(self.local_config_file_path):
            Logger.warn("The working directory has already been initialized.")

        config = get_init_config()
        
        ### use init_config instead of local/inherited config
        if custom_init_config:
            merge_dicts(custom_init_config, config)
            ### merge with existing local configuration?
            # if override_inherited:
            #     Logger.debug("Merging local configuration...")
            #     self.load_local_config(silent_warnings=True, render=False)
            #     merge_dicts(self.local_config, config)
            Logger.debug("Merging exisintg configuration...")
            self.load_local_config(silent_warnings=True, render=False)
            merge_dicts(self.local_config, config)
        else:
            ### override locally inherited configuration?
            if override_inherited: 
                Logger.debug("Merging root configuration...")
                self.load_root_config(render=False)
                merge_dicts(self.root_config, config)

                # Logger.debug("Merging inherited configuration...")
                # self.load_inherited_config(render=False)
                # merge_dicts(self.inherited_config, config)
            
            ### do not show warning if directory is not initialized yet
            Logger.debug("Merging existing configuration...")
            self.load_local_config(silent_warnings=True, render=False)
            merge_dicts(self.local_config, config)

        ### if config overrides, merge them to local
        if config_overrides_string:
            Logger.debug("Merging configuration overrides...")
            self.apply_config_overrides(config_overrides_string)
            merge_dicts(self.overriden_config, config)

        self.local_config = config
        self.update_merged_config()

        self.save_local_config()


    def list_local(self, filter=None):
        self.update_merged_config(use_remote=False) 
        response = flatten_dict(self.merged_config)
        result = []
        for key, value in response.items():
            if filter and not re.match(filter, key): continue
            item = {
                'Key' : key,
                'Value': value,
                'Context': 'local',
                'Path': os.path.join(self.config_dir, self.config_file)
            }
            result.append(item)

        return result
    
    def list_remote(self, uninherited=False, filter=None):
        if self.config_provider:
            response = RemoteConfig.list(uninherited=uninherited, filter=filter)
        else:
            Logger.warn("Remote configiguration is not availbale because config provider has not been set.")
            response = []
        return response


    def list(self, uninherited=False, filter=None, rendered=False, source=ConfigOrigin.All):
        Logger.info(f"Listing configuration settings: namespace={self.get_namespace(ContextMode.Target)}, application={self.get_application(ContextMode.Target)}, stage={self.get_stage(ContextMode.Target)}, scope={self.scope}")
        if source == ConfigOrigin.Local:  
            response = self.list_local(filter=filter)
        elif source == ConfigOrigin.Remote:
            response = self.list_remote(uninherited=uninherited, filter=filter)
        elif source == ConfigOrigin.All:
            local_response = self.list_local(filter=filter)
            response = self.list_remote(uninherited=uninherited, filter=filter)
            key_index = {response[i]['Key']:i for i in range(len(response))}
            for item in local_response:
                if item['Key'] in key_index:
                    response[key_index[item['Key']]] = item
                    Logger.warn(f"'{item['Key']}' was overriden locally.")
                else:
                    response.append(item)

        if rendered:
            self.load_remote_config()
            self.load_root_config()
            self.load_local_config()
            response = render_dict_values(response, self.merged_config)

        from operator import itemgetter
        return {'Configuration': sorted(response, key=itemgetter('Key'))}


    def set(self, key, value, source=ConfigOrigin.Local):
        Logger.info(f"Adding configuration setting '{key}': namespace={self.get_namespace(ContextMode.Target)}, application={self.get_application(ContextMode.Target)}, stage={self.get_stage(ContextMode.Target)}, scope={self.scope}")
        self.validate_key(key, "get")
        if source == ConfigOrigin.Local:
            # if not os.path.exists(self.local_config_file_path):
            #     raise DSOException("The working directory has not been intitialized yet. Run 'dso config init' to do so.")
            value, key = set_item(self.local_config, key.split('.'), value, overwrite_parent=False, overwrite_children=False)
            self.save_local_config()
            return {
                'Key' : '.'.join(key),
                'Value' : value,
                'Context': 'local',
                'Path': os.path.join(self.config_dir, self.config_file)
            }            
            # self.load_local_config(render=False)
        elif source == ConfigOrigin.Remote:
            return RemoteConfig.set(key=key, value=value)
        else:
            raise NotImplementedError()



    def get_local(self, key):
        self.update_merged_config(use_remote=False)
        item, key = get_item(self.merged_config, key.split('.'), create=False, leaf_only=True)
        if item is None:
            return {}
        return {
            'Key' : '.'.join(key),
            'Value' : item.copy() if type(item) in [list,dict] else item,
            'Context': 'local',
            'Path': os.path.join(self.config_dir, self.config_file)
        }

    def get_remote(self, key, revision, uninherited=False):
        ### are we dealing with a list?
        if '.' in key:
            lastKey = key.split('.')[-1]
            if is_int(lastKey):
                lastKey = int(lastKey)
                key = '.'.join(key.split('.')[:-1])    
                result = RemoteConfig.get(key=key, revision=revision, uninherited=uninherited)
                if type(result['Value']) == list:
                    if lastKey > len(result['Value']):
                        raise DSOException(f"Index '{lastKey}' exceeded the last index of list: {len(result['Value'])-1}")
                    
                    result['Key'] += f'.{lastKey}'
                    result['Value'] = result['Value'][lastKey]
                else:
                    raise DSOException(f"Index qualifiers can only be used with lists not with {type(result['Value'])}.")
            else:
                result = RemoteConfig.get(key=key, revision=revision, uninherited=uninherited)
                
        if not result:
            raise DSOException(f"Configuration setting '{key}' not found in the given context: namespace={self.get_namespace(ContextMode.Target)}, application={self.get_application(ContextMode.Target)}, stage={self.get_stage(ContextMode.Target)}, scope={self.scope}")
        return result

    ### handle only sinple lists, not list of dicts
    def is_parameterized(self, data):
        if type(data) == str:
            return re.match(r'.*{{.+}}.*', data)
        try:
            iterator = iter(data)
            for item in iterator:
                if re.match(r'.*{{.+}}.*', str(item)):
                    return True
        except TypeError:
            return re.match(r'.*{{.+}}.*', str(data))
        return False


    def get(self, key, revision=None, uninherited=False, rendered=True, source=ConfigOrigin.All):
        Logger.info(f"Getting configuration setting '{key}': namespace={self.get_namespace(ContextMode.Target)}, application={self.get_application(ContextMode.Target)}, stage={self.get_stage(ContextMode.Target)}, scope={self.scope}")
        self.validate_key(key, "get")
        if source == ConfigOrigin.All:
            result = self.get_local(key)
            if not result:
                Logger.debug(f"Configuration setting '{key}' not found locally.")   
                if self.config_provider:
                    result = self.get_remote(key=key, revision=revision, uninherited=uninherited)
                    if not result:
                        raise DSOException(f"Configuration setting '{key}' not found in the given context: namespace={self.get_namespace(ContextMode.Target)}, application={self.get_application(ContextMode.Target)}, stage={self.get_stage(ContextMode.Target)}, scope={self.scope}")
                else:
                    Logger.warn("Remote configiguration is not availbale because config provider has not been set.")
                    result = {}
        elif source == ConfigOrigin.Local:
            result = self.get_local(key)
            if not result:
                raise DSOException(f"Configuration setting '{key}' not found locally.")   
        elif source == ConfigOrigin.Remote:
            result = self.get_remote(key=key, revision=revision, uninherited=uninherited)

        if rendered and result and self.is_parameterized(result['Value']):
            ### load everything for rendering purpose only
            self.load_remote_config()
            self.load_root_config()
            self.load_local_config()
            self.update_merged_config()
            result['Value'] = render_dict_values(result['Value'], values=self.merged_config, silent=False)
        
        return result
            

    def delete_remote(self, key):
        return RemoteConfig.delete(key=key)


    def delete_local(self, key):
        parent = get_item(self.local_config, key.split('.')[:-1])[0]
        lastKey = key.split('.')[-1]
        if parent and type(parent) in [dict, list] and lastKey:
            value = get_item(data=self.local_config, keys=key.split('.'), create=False, leaf_only=True)[0]
            if value is None:
                raise DSOException(f"'{key}' not found in configuration settings locally.")
            del_item(dic=self.local_config, keys=key.split('.'), leaf_only=True, silent=False)
            del_dict_empty_item(dic=self.local_config, keys=key.split('.')[:-1])
            self.save_local_config()
            self.load_local_config(render=False)
        else:
            raise DSOException(f"'{key}' not found in configuration settings locally.")
        
        return {
            'Key' : key,
            'Value': value,
            'Context': 'local',
            'Path': os.path.join(self.config_dir, self.config_file)
        }   
        

    def unset(self, key, source=ConfigOrigin.Local):
        Logger.info(f"Deleting configuration setting '{key}': namespace={self.get_namespace(ContextMode.Target)}, application={self.get_application(ContextMode.Target)}, stage={self.get_stage(ContextMode.Target)}, scope={self.scope}")
        self.validate_key(key, "get")
        if source == ConfigOrigin.Local:
            return self.delete_local(key)
        elif source == ConfigOrigin.Remote:
            return self.delete_remote(key)
        else:
            raise NotImplementedError()
  

    def history_local(self, key):
        self.update_merged_config(use_remote=False)
        response = get_item(self.merged_config, key.split('.'), create=False, leaf_only=True)[0]
        if response:
            return {'Revisions': [{
                'RevisionId': '0',
                'Key' : key,
                'Value' : response,
                'Context': 'local',
                'Path': os.path.join(self.config_dir, self.config_file),
                'Date': get_file_modified_date(os.path.join(self.config_dir, self.config_file)),
            }]}
        else:
            raise DSOException(f"Configuration setting '{key}' not found locally.")


    def history_remote(self, key):
        result = RemoteConfig.history(key)
        if not result:
            raise DSOException(f"Configuration setting '{key}' not found in the given context: namespace={self.get_namespace(ContextMode.Target)}, application={self.get_application(ContextMode.Target)}, stage={self.get_stage(ContextMode.Target)}, scope={self.scope}")
        return result


    def history(self, key, source=ConfigOrigin.Remote):
        Logger.info(f"Fetching history of configuration setting '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        self.validate_key(key, "get")
        if source == ConfigOrigin.Local:
            Logger.warn(f"Local configuration does not support history.")
            result = self.history_local(key)
        elif source == ConfigOrigin.Remote:
            return self.history_remote(key)
        else:
            raise NotImplementedError()


        return result


Config = ConfigService()




from dsocli.exceptions import DSOException
from dsocli.logger import Logger
from dsocli.providers import Providers
from dsocli.releases import ReleaseProvider
from dsocli.constants import *
from dsocli.contexts import ContextMode
from dsocli.aws_s3_utils import *
from dsocli.settings import *
from dsocli.configs import Config


__default_spec = {
    'foo': 'bar',
}


def get_default_spec():
    return __default_spec.copy()


class AwsCfnReleaseProvider(ReleaseProvider):

    def __init__(self):
        super().__init__('release/aws/cfn/v1')

    def create(self):
        Logger.debug(f"Building CFN release: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        return 'release1.zip'


def register():
    Providers.register(AwsCfnReleaseProvider())



import os
import imp
from .configs import Config
from .constants import *
from .logger import Logger
from .exceptions import DSOException

class ProviderBase():
    def __init__(self, id):
        self.__id = id
    @property
    def id(self):
        return self.__id

class KeyValueStoreProvider(ProviderBase):

    # def validate_key(self, key):
    #     Logger.debug(f"Validating: key={key}")
    #     pattern = self.get_key_validator()
    #     if not re.match(pattern, key):
    #         raise DSOException(MESSAGES['InvalidKeyPattern'].format(key, pattern))

    # def get_key_validator(self, key):
    #     raise NotImplementedError()

    def list(self):
        raise NotImplementedError()
    def add(self):
        raise NotImplementedError()
    def delete(self):
        raise NotImplementedError()
    def get(self):
        raise NotImplementedError()
    def history(self):
        raise NotImplementedError()


class ArtifactoryProvider(ProviderBase):
    def list(self):
        raise NotImplementedError()
    def upload(self):
        raise NotImplementedError()
    def download(self):
        raise NotImplementedError()
    def history(self):
        raise NotImplementedError()
    def delete(self):
        raise NotImplementedError()


class RemoteConfigProvider(KeyValueStoreProvider):
    def list(self, filter=None, uninherited=False):
        raise NotImplementedError()
    def add(self, key, value):
        raise NotImplementedError()
    def get(self, key, revision=None, uninherited=False, rendered=True):
        raise NotImplementedError()
    def history(self, key):
        raise NotImplementedError()
    def delete(self, key):
        raise NotImplementedError()


class ProviderService():
    __providers = {}

    # def load_all_providers(self):
    #     __import__(AppConfigsroot_path + 'lib/dso/provider')
    #     # importdir.do(os.path.dirname(__file__)+'/secret_providers', globals())
    #     # importdir.do(os.path.dirname(__file__)+'/template_providers', globals())

    def load_provider(self, provider_slug):
        Logger.debug(f"Loading provider '{provider_slug}'...")
        providerPackagePath = os.path.join(Config.install_path, 'provider', provider_slug)
        if not os.path.exists(providerPackagePath):
            raise DSOException(f"Provider '{provider_slug}' not found.")
        imp.load_package(provider_slug, providerPackagePath).register()

    def register(self, provider: ProviderBase):
        if not provider.id in self.__providers:
            self.__providers[provider.id] = provider
            Logger.debug(f"Provider registered: id={provider.id}")

    def get_provider(self, provider_slug):
        if not provider_slug in self.__providers:
            self.load_provider(provider_slug)

        ### make sure provider has registered, and return it
        if provider_slug in self.__providers:
            return self.__providers[provider_slug] 
        else:
            raise DSOException(f"No provider has registered for '{provider_slug}'.")

    def RemoteConfigProvider(self):
        if not Config.config_provider:
            raise DSOException(MESSAGES['ProviderNotSet'].format('Config'))
        return self.get_provider('config/' + Config.config_provider)

    def ParameterProvider(self):
        if not Config.parameter_provider:
            raise DSOException(MESSAGES['ProviderNotSet'].format('Parameter'))
        return self.get_provider('parameter/' + Config.parameter_provider)

    def SecretProvider(self):
        if not Config.secret_provider:
            raise DSOException(MESSAGES['ProviderNotSet'].format('Secret'))
        return self.get_provider('secret/' + Config.secret_provider)

    def TemplateProvider(self):
        if not Config.template_provider:
            raise DSOException(MESSAGES['ProviderNotSet'].format('Template'))
        return self.get_provider('template/' + Config.template_provider)

    def ArtifactoryProvider(self):
        if not Config.artifactory_provider:
            raise DSOException(MESSAGES['ProviderNotSet'].format('Artifactory'))
        return self.get_provider('artifactory/' + Config.artifactory_provider)

    def PackageProvider(self):
        if not Config.package_provider:
            raise DSOException(MESSAGES['ProviderNotSet'].format('Package'))
        return self.get_provider('package/' + Config.package_provider)

    def ReleaseProvider(self):
        if not Config.release_provider:
            raise DSOException(MESSAGES['ProviderNotSet'].format('Release'))
        return self.get_provider('release/' + Config.release_provider)

Providers = ProviderService()



import os
from dsocli.logger import Logger
from dsocli.configs import Config, ContextMode
from dsocli.providers import Providers, RemoteConfigProvider
from dsocli.stages import Stages
from dsocli.constants import *
from dsocli.exceptions import DSOException
from dsocli.local_utils import *
from dsocli.settings import *
from jinja2.filters import K


__default_spec = {
    'path': os.path.join(Config.config_dir, 'config/'),
    'store': 'local.json',
}


def get_default_spec():
    return __default_spec.copy()



class LocalConfigProvider(RemoteConfigProvider):

    def __init__(self):
        super().__init__('config/local/v1')

    @property
    def root_path(self):
        return Config.config_spec('path')


    def get_path_prefix(self):
        return self.root_path



    @property
    def store_name(self):
        return Config.config_spec('store')



    def set(self, key, value):
        Logger.debug(f"Adding local configuration setting '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        response = add_local_parameter(key=key, value=value, store_name=self.store_name, path_prefix=self.get_path_prefix())
        return response



    def list(self, uninherited=False, filter=None):
        Logger.debug(f"Listing local configuration settings: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        settings = load_context_local_parameters(store_name=self.store_name, path_prefix=self.get_path_prefix(), uninherited=uninherited, filter=filter)
        result = []
        for key, details in settings.items():
            item = {
                'Key': key,
            }
            item.update(details)
            result.append(item)

        return result



    def get(self, key, revision=None, uninherited=False, rendered=True):
        if revision:
            Logger.warn(f"Config provider 'local/v1' does not support versioning. Revision request ignored.")
        Logger.debug(f"Getting configuration setting '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        found = locate_parameter_in_context_hierachy(key=key, store_name=self.store_name, path_prefix=self.get_path_prefix(), uninherited=uninherited)
        if not found:
            raise DSOException(f"Configuration setting '{key}' not found in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        # if len(found) > 1:
        #     raise DSOException(f"Mutiple settings found with the same key '{key}' in the given context.")
        result = {
                'Key': key, 
            }
        result.update(found[key])
        return result



    def history(self, key):
        Logger.warn(f"Config provider 'local/v1' does not support history.")

        Logger.debug(f"Getting local configuration setting '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        found = locate_parameter_in_context_hierachy(key=key, store_name=self.store_name, path_prefix=self.get_path_prefix(), uninherited=False)
        if not found:
            raise DSOException(f"Configuration setting '{key}' not found in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        result = { "Revisions":
            [{
                'RevisionId': '0',
                'Date': get_file_modified_date(found[key]['Path']),
            }]
        }
        result['Revisions'][0].update(found[key])
        return result


    def unset(self, key):
        Logger.debug(f"Locating config setting '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        ### only configs owned by the config can be deleted, hence uninherited=True
        found = locate_parameter_in_context_hierachy(key=key, store_name=self.store_name, path_prefix=self.get_path_prefix(), uninherited=True)
        if not found:
            raise DSOException(f"Configuration setting '{key}' not found in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        Logger.info(f"Deleting configuration setting '{key}': path={found[key]['Path']}")
        delete_local_parameter(found[key]['Path'], key=key)
        result = {
                'Key': key,
                'Value': found[key]['Value'],
                'Stage': Config.short_stage,
                'Scope': found[key]['Scope'], 
                'Context': found[key]['Context'], 
                'Path': found[key]['Path'],
            }
        return result



def register():
    Providers.register(LocalConfigProvider())



import boto3
from dsocli.exceptions import DSOException
from dsocli.logger import Logger
from dsocli.providers import Providers, RemoteConfigProvider
from dsocli.constants import *
from dsocli.aws_ssm_utils import *
from dsocli.configs import Config, ContextMode


__default_spec = {
    'pathPrefix': '/dso/config/',
}

def get_default_spec():
    return __default_spec.copy()


class AwsSsmConfigProvider(RemoteConfigProvider):

    def __init__(self):
        super().__init__('config/aws/ssm/v1')


    ### adds service name to the artifactory prefix
    def get_path_prefix(self):
        return Config.config_spec('pathPrefix')


    def list(self, uninherited=False, filter=None):
        Logger.debug(f"Listing SSM configuration settings: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        configuration = load_context_ssm_parameters(parameter_type='String', path_prefix=self.get_path_prefix(), uninherited=uninherited, filter=filter)
        result = []
        for key, details in configuration.items():
            item = {
                'Key': key,
                'RevisionId': str(details['Version']),
            }
            item.update(details)
            result.append(item)

        return result


    def set(self, key, value):
        Logger.debug(f"Checking configuration overwrites '{key}'...")
        assert_ssm_parameter_no_namespace_overwrites(key=key, path_prefix=self.get_path_prefix())
        Logger.debug(f"Locating configuration setting '{key}'...")
        found = locate_ssm_parameter_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(), uninherited=True)
        if found and not found['Type'] == 'String':
            raise DSOException(f"Failed to add configuration setting '{key}' becasue the key is not available in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        path = get_ssm_path(context=Config.context, key=key, path_prefix=self.get_path_prefix())
        Logger.debug(f"Adding configuration setting: path={path}")
        response = add_ssm_paramater(path, value)
        result = {
                'RevisionId': str(response['Version']),
                'Key': key, 
                'Value': value,
                'Stage': Config.short_stage,
                'Scope': Config.context.scope_translation,
                'Context': {
                    'Namespace': Config.namespace,
                    'Application': Config.application,
                    'Stage': Config.stage,
                },
                'Path': path,
            }
        result.update(response)
        return result



    def get(self, key, revision=None, uninherited=False, rendered=True):
        Logger.debug(f"Locating configuration setting '{key}'...")
        found = locate_ssm_parameter_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(), uninherited=uninherited)
        if not found:
            raise DSOException(f"Config '{key}' not found in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        else:
            if not found['Type'] == 'String':
                raise DSOException(f"Configuration setting '{key}' not found in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        Logger.debug(f"Getting configuration setting from SSM: path={found['Name']}")
        response = get_ssm_parameter_history(found['Name'])
        configuration = sorted(response['Parameters'], key=lambda x: int(x['Version']), reverse=True)
        if revision is None:
            ### get the latest revision
            result = {
                    'RevisionId': str(configuration[0]['Version']),
                    'Date': configuration[0]['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key, 
                    'Value': configuration[0]['Value'],
                    'Scope': found['Scope'],
                    'Context': found['Context'],
                    'Path': found['Name'],
                    'User': configuration[0]['LastModifiedUser'],
                    }
                
        else:
            ### get specific revision
            configuration = [x for x in configuration if str(x['Version']) == revision]
            if not configuration:
                raise DSOException(f"Revision '{revision}' not found for configuration setting '{key}' in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
            result = {
                    'RevisionId':str(configuration[0]['Version']),
                    'Date': configuration[0]['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                    'Key': key, 
                    'Value': configuration[0]['Value'],
                    'Scope': found['Scope'],
                    'Context': found['Context'],
                    'Path': found['Name'],
                    'User': configuration[0]['LastModifiedUser'],
                    }

        return result


    def history(self, key):
        Logger.debug(f"Locating configuration setting '{key}'...")
        found = locate_ssm_parameter_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(), uninherited=True)
        if not found:
            raise DSOException(f"Config '{key}' not found in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        else:
            if not found['Type'] == 'String':
                raise DSOException(f"Config '{key}' not found in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        Logger.debug(f"Getting configuration from SSM: path={found['Name']}")
        response = get_ssm_parameter_history(found['Name'])
        configuration = sorted(response['Parameters'], key=lambda x: int(x['Version']), reverse=True)
        result = { "Revisions":
            [{
                'RevisionId': str(setting['Version']),
                'Date': setting['LastModifiedDate'].strftime('%Y/%m/%d-%H:%M:%S'),
                'Key': key,
                'Value': setting['Value'],
                # 'Scope': found['Scope'],
                # 'Context': found['Context'],
                'User': setting['LastModifiedUser'],
                # 'Path': found['Name'],
            } for setting in configuration]
        }

        return result



    def unset(self, key):
        Logger.debug(f"Locating configuration setting '{key}'...")
        ### only configuration owned by the context can be deleted, hence uninherited=True
        found = locate_ssm_parameter_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(), uninherited=True)
        if not found:
            raise DSOException(f"Configuration '{key}' not found in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        else:
            if not found['Type'] == 'String':
                raise DSOException(f"Configuraton '{key}' not found in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        Logger.debug(f"Deleting configuration from SSM: path={found['Name']}")
        delete_ssm_parameter(found['Name'])
        result = {
                'Key': key,
                'Value': found['Value'],
                'Stage': found['Stage'],
                'Scope': found['Scope'],
                'Context': found['Context'],
                'Path': found['Name'],
                }
        return result


def register():
    Providers.register(AwsSsmConfigProvider())




from .constants import *
from .providers import ProviderBase, Providers
from .artifacts import Artifactory
from .logger import Logger
from .exceptions import DSOException
from .configs import Config


key_regex_pattern = r"^[a-zA-Z]([./a-zA-Z0-9_-]*[a-zA-Z0-9])?$"

class PackageProvider(ProviderBase):
    def list(self, filter=None):
        raise NotImplementedError()
    def build(self, key):
        raise NotImplementedError()
    def get(self, key, revision=None):
        raise NotImplementedError()
    # def history(self, key):
    #     raise NotImplementedError()
    def delete(self, key):
        raise NotImplementedError()


class PackageService():
    

    service_name = 'package'


    # def validate_key(self, key):
    #     Logger.info(f"Validating package key '{key}'...")
    #     if not key:
    #         raise DSOException(MESSAGES['KeyNull'])
    #     if key == 'dso' or key.startswith('dso.'):
    #         raise DSOException(MESSAGES['DSOReserverdKey'].format(key))
    #     if not re.match(key_regex_pattern, key):
    #         raise DSOException(MESSAGES['InvalidKeyPattern'].format(key, key_regex_pattern))
    #     ### the regex does not check adjacent special chars
    #     if '..' in key:
    #         raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '..'))

    #     if '//' in key:
    #         raise DSOException(MESSAGES['InvalidKeyStr'].format(key, '//'))


    @property
    def version_major(self):
        try:
            return int(Config.get(key='version.major', service=self.service_name)['Value'])
        except DSOException:
            init_version_major = 0
            self.version_major = init_version_major
            return init_version_major

    @version_major.setter
    def version_major(self, value):
        Config.set(key='version.major', value=str(value), service=self.service_name)


    @property
    def version_minor(self):
        try:
            return int(Config.get(key='version.minor', service=self.service_name)['Value'])
        except DSOException:
            init_version_minor = 0
            self.version_minor = init_version_minor
            return init_version_minor

    @version_minor.setter
    def version_minor(self, value):
        Config.set(key='version.minor', value=str(value), service=self.service_name)


    @property
    def version_patch(self):
        try:
            return int(Config.get(key='version.patch', service=self.service_name)['Value'])
        except DSOException:
            init_version_patch = 0
            self.version_patch = init_version_patch
            return init_version_patch

    @version_patch.setter
    def version_patch(self, value):
        Config.set(key='version.patch', value=str(value), service=self.service_name)


    @property
    def version_build(self):
        try:
            return int(Config.get(key='version.build', service=self.service_name)['Value'])
        except DSOException:
            init_version_build = 0
            self.version_build = init_version_build
            return init_version_build


    @version_build.setter
    def version_build(self, value):
        Config.set(key='version.build', value=str(value), service=self.service_name)


    def list(self, filter=None):
        provider = Providers.PackageProvider()
        Logger.info(f"Listing packages: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        response = Artifactory.list(service=self.service_name, filter=filter)
        result = {'Packages': response['Artifacts']}
        return result


    def build(self):
        provider = Providers.PackageProvider()
        Logger.info(f"Building package: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        artifact = provider.build()
        major = self.version_major
        minor = self.version_minor
        patch = self.version_patch
        build = self.version_build
        artifactKey = f"{major}.{minor}.{patch}.{build}"
        self.version_build = build + 1
        Logger.info(f"Adding package '{artifactKey}' to artifact store...")
        # changed = Logger.decrease_verbosity()
        # try:
        #     response = Artifactory.add(filepath=artifact, key=artifactKey ,service=self.service_name)
        # finally:
        #     if changed: Logger.increase_verbosity()
        response = Artifactory.add(filepath=artifact, key=artifactKey ,service=self.service_name)
        return response


    def get(self, key):
        Logger.info(f"Getting package '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        response = Artifactory.get(key=key, service=self.service_name)
        return response


    # def history(self, key):
    #     self.config = config
    #     provider = Providers.PackageProvider()
    #     Logger.info(f"Fetching history of package '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
    #     return provider.history(key)


    def delete(self, key):
        Logger.info(f"Deleting package '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        response = Artifactory.delete(key=key, service=self.service_name)
        return response


Packages = PackageService()


import re
from unittest import result
from venv import create
from .logger import Logger
from .exceptions import DSOException



# def merge_dicts(source, destination):
#     if not source: return destination
#     for key, value in source.items():
#         if isinstance(value, dict):
#             if key in destination.keys():
#                 if not isinstance(destination[key], dict):
#                     raise Exception(f"Failed to merge '{key}' beacuse destination has an existing key with incompatible type ({type(destination[key])}) to that of the source ({type(source[key])}).")
#             else:
#                 destination[key] = {}
#             node = destination[key]
#             merge_dicts(value, node)
#         else:
#             destination[key] = value

#     return destination


# def merge_dicts(source: dict, destination: dict, merge_key=None):
    
#     if not isinstance(source, dict) or not isinstance(destination, dict):
#         raise Exception("Failed to merge, 'dict' type is expected for source and destination.")

#     def find_item(item, alist, merge_key=None):
#         if merge_key:
#             if callable(merge_key):
#                 found = [x for x in alist if merge_key(x) == merge_key(item)]
#             else:
#                 if isinstance(item, dict) and merge_key in item:
#                     found = [x for x in alist if x[merge_key] == item[merge_key]]
#                 else:
#                     found = [x for x in alist if x == item]
#         else:
#             found = [x for x in alist if x == item]

#         return found[0] if found else None

#     if not source: return destination
    
#     for key, value in source.items():
#         if isinstance(value, dict):
#             if key in destination.keys():
#                 if not isinstance(destination[key], dict):
#                     raise Exception(f"Failed to merge '{key}' beacuse destination has an existing key with incompatible type ({type(destination[key])}) to that of the source ({type(source[key])}).")
#             else:
#                 destination[key] = {}
#             node = destination[key]
#             merge_dicts(value, node, merge_key)
#         elif isinstance(value, list):
#             if key in destination.keys():
#                 if not isinstance(destination[key], list):
#                     raise Exception(f"Failed to merge '{key}' beacuse destination has an existing key with incompatible type ({type(destination[key])}) to that of the source ({type(source[key])}).")
#             else:
#                 destination[key] = []
#             for item in value:
#                 if not isinstance(item, dict):

#                 node = find_item(item, destination[key], merge_key)
#                 if node:
#                     if isinstance(node, dict):
#                         if not isinstance(item, dict):
#                             raise Exception("Failed to merge. Destination has an existing key with incompatible type.")
#                         merge_dicts(item, node, merge_key)
#                     else:
#                         if destination[key].index(item) < 0:
#                             destination[key].append(item)
#                 else:
#                     destination[key].append(item)
#         # elif isinstance(value, tuple):
#         #     raise NotImplementedError
#         # elif isinstance(value, set):
#         #     raise NotImplementedError
#         else:
#             destination[key] = value

#     return destination


def merge_dicts(source, destination, merge_key=None):
    
    def find_item(item, alist, merge_key=None):
        found = None
        if merge_key:
            if callable(merge_key):
                for x in alist:
                    if merge_key(x) == merge_key(item):
                        found = x
                        break
                # found = [x for x in alist if merge_key(x) == merge_key(item)]
            else:
                if isinstance(item, dict) and merge_key in item:
                    for x in alist:
                        if isinstance(x, dict) and merge_key in x and x[merge_key] == item[merge_key]:
                            found = x
                            break
                    # found = [x for x in alist if merge_key in x and x[merge_key] == item[merge_key]]
                else:
                    for x in alist:
                        if x == item:
                            found = x
                            break
                    # found = [x for x in alist if x == item]
        else:
            for x in alist:
                if x == item:
                    found = x
                    break

        return found

    if source is None: 
        return destination

    # if destination is None:
    #     destination = type(source)()

    if not type(source) == type(destination):
        raise Exception(f"Failed to merge source ({source}) and destination ({destination}) due to incompatible types: {type(source)} and {type(destination)}")

    if isinstance(source, dict):
        for key, value in source.items():
            if isinstance(value, dict) or isinstance(value, list):
                if not key in destination.keys(): destination[key] = type(value)()
                node = destination[key]
                merge_dicts(value, node, merge_key)
            else:
                destination[key] = value

    elif isinstance(source, list):
        for item in source:
            node = find_item(item, destination, merge_key)
            if node:
                if isinstance(node, dict) or isinstance(node, list):
                    merge_dicts(item, node, merge_key)
            else:
                destination.append(item)
    else:
        raise NotImplementedError


    return destination



def flatten_dict(input, prefixed_key = '', delimiter = '.', atomic_list=True):
    def visit(input, prefixed_key, delimiter, output):
        if isinstance(input, dict):
            for key, value in input.items():
                new_key = f'{prefixed_key}{delimiter}{key}' if prefixed_key else f'{key}'
                visit(value, new_key, delimiter, output)
        elif isinstance(input, list):
            if atomic_list:
                output[prefixed_key] = input        
            else:
                for idx, item in enumerate(input):
                    new_key = f'{prefixed_key}{delimiter}{idx}' if prefixed_key else f'{idx}'
                    visit(item, new_key, delimiter, output)
        else:
            output[prefixed_key] = input

        return output

    return visit(input, prefixed_key, delimiter, {})



def deflatten_dict(input: dict, delimiter = '.'):
    data = {}
    for key, value in input.items():
        set_item(data, key.split(delimiter), value, overwrite_parent=True, overwrite_children=True)
    return data



# def get_item(data, keys, create=True, default={}, leaf_only=False):
#     for i in range(len(keys)):
#         key = keys[i]
#         if isinstance(data, dict):
#             if not key in data.keys() or data[key] is None:
#                 if create:
#                     data[key] = default.copy() ### !Important
#                 else:
#                     return None
#             data = data[key]
#         elif isinstance(data, list):
#             if key == '*':
#                 key = len(data)

#             key = int(key)
#             if key >= len(data):
#                 raise DSOException(f"Index '{key}' exceeded list size: {'.'.join(keys[:-1])}.{len(data)-1}")
#             data = data[key]
#         else:
#             raise NotImplementedError()
#     if leaf_only and type(data) in [dict]:
#         return None
#     return data
    


def safe_str_to_number(s):
    try:
        return int(s.strip())
    except:
        try:
            return float(s.strip())
        except:
            return s
            

def is_number(s):
    try:
        return type(int(s.strip())) == int
    except:
        try:
            return type(float(s.strip())) == float
        except:
            return False
            

def is_int(s):
    try:
        return type(int(s.strip())) == int
    except:
        return False
            

def get_item(data, keys, create=True, default={}, leaf_only=False):
    def create_new_item(keys, default):
        ### is this a leaf item?
        if len(keys) == 1:
            return default.copy() ### !Important
            # return {}.copy() ### !Important
        else:
            ### decide type of the new item based on the next key
            if keys[1] == '*' or is_int(keys[1]):
                return [].copy() ### !Important
            else:
                return {}.copy() ### !Important

    def get_next_item(parent, keys, create, default):
        if type(parent) == dict:
            if not keys[0] in parent.keys() or parent[keys[0]] is None:
                if create:
                    parent[keys[0]] = create_new_item(keys, default)
                else:
                    return None, keys
            return parent[keys[0]], keys
        elif type(parent) == list:
            if keys[0] == '*':
                ### '*' is allowed only when creating/appending is enabled
                if not create:
                    return None, keys
                keys[0] = str(len(parent))
                parent.append(create_new_item(keys, default))
            else:
                if not is_int(keys[0]):
                    Logger.warn(f"Encountered a list but invalid index qualifier '{keys[0]}' provided: {parent}")
                    return None, keys
            
            if int(keys[0]) >= len(parent):
                raise DSOException(f"Index '{keys[0]}' exceeded the last index of list: {len(parent)-1}")

            return parent[int(keys[0])], keys

    if len(keys) == 0 or data is None:
        return data, keys
    if len(keys) == 1:
        if not type(data) in [dict, list]:
            return None, keys
        item, keys = get_next_item(data, keys, create, default)
    else:
        parent, keys = get_next_item(data, keys, create, default)
        item, keys2 = get_item(parent, keys[1:], create, default)
    
    if leaf_only and type(item) == dict:
        return None, keys
    else:
        return item, keys



def set_item(dic, keys, value, overwrite_parent=False, overwrite_children=False):
    
    def set_list_value(alist, idx, value):
        idx = int(idx)
        if idx < len(alist):
            if type(alist[idx]) == dict:
                if overwrite_children:
                    Logger.warn("'{0}' was overwritten.".format('.'.join(keys)))
                else:
                    raise DSOException(f"DSO did not set '{'.'.join(keys)}' becasue it would otherwise overwrite an existing item '{'.'.join(keys)}' of type {type(alist)}`.")
            alist[idx] = value
        elif idx == len(alist):
            alist.append(value)
        else:
            raise DSOException(f"Index '{idx}' exceeded the last index of list: {len(alist)-1}")
    
        return idx

    if not keys:
        return
        
    lastKey = keys[-1]
    parent, keys = get_item(dic, keys[:-1], create=True, default={}, leaf_only=False)
    if lastKey == '*':
        if not isinstance(parent, list):
            raise DSOException(f"Index qualifiers can only be used with lists not {type(parent)}.")
        ### append to the list
        lastKey = len(parent)
    elif is_int(lastKey):
        lastKey = int(lastKey)

    keys.append(str(lastKey))


    ### is value a comma separated list enclosed in brackets?
    if type(value) == str:
        if re.match(r'^\[(.*)\]$', value):
            value = re.sub(r"^\[", "", re.sub(r"\]$", "", value))
            value = list(map(lambda x: safe_str_to_number(x.strip()), re.findall('([^,]+)', value)))
            ### alow overrwrite the entire list
            if not lastKey in parent or parent[lastKey] is None or not type(parent[lastKey]) in [dict]:
                overwrite_children = True
            else:
                raise DSOException(f"DSO did not set '{'.'.join(keys)}' becasue it would otherwise overwrite an existing item of type {type(parent[lastKey])}`.")
        else:
            value = safe_str_to_number(value)


    if isinstance(parent, list):
        set_list_value(parent, lastKey, value)
    elif isinstance(parent, dict):
        if lastKey in parent.keys():
            ### item is expected to be a basic type (string, number, ...)
            if type(parent[lastKey]) == dict:
                if overwrite_children:
                    Logger.warn("'{0}' was overwritten.".format('.'.join(keys)))
                else:
                    raise DSOException(f"DSO did not set '{'.'.join(keys)}' becasue it would otherwise overwrite an existing item of type {type(parent[lastKey])}`.")
        parent[lastKey] = value
    else:
        if overwrite_parent or parent is None:
            grand_parent = get_item(dic, keys[:-2])[0]
            ### are we dealing with a list or a dict?
            if type(lastKey) == int:
                grand_parent[keys[-2]] = []
                parent = grand_parent[keys[-2]]
                set_item(parent, lastKey, value)
            else:
                grand_parent[keys[-2]] = {}
                parent = grand_parent[keys[-2]]
                parent[lastKey] = value
        else:
            raise DSOException(f"DSO did not set '{'.'.join(keys)}' becasue it would otherwise overwrite an existing item '{'.'.join(keys[:-1])}' of type {type(parent)}.")

    return value, keys


def del_item(dic, keys, leaf_only=False, silent=False):
    parent = get_item(dic, keys[:-1], create=False)[0]
    if not parent:
        return False
    if type(parent) == dict:
        if not keys[-1] in parent:
            return False
        item = parent[keys[-1]]
        if isinstance(item, dict):
            if leaf_only:
                if not silent:
                    raise DSOException("'{0}' is a dictionary and cannot be deleted.".format('.'.join(keys)))
                return False

        parent.pop(keys[-1])
        return True

    elif type(parent) == list:
        if int(keys[-1]) >= len(parent):
            return False
        item =  parent[int(keys[-1])]

        if isinstance(item, dict):
            if leaf_only:
                if not silent:
                    raise DSOException("'{0}' is a dictionary and cannot be deleted.".format('.'.join(keys)))
                return False

        parent.pop(int(keys[-1]))
        return True
    else:
        raise NotImplementedError()



def del_dict_empty_item(dic, keys):
    item = get_item(dic, keys)[0]
    if not item:
        del_item(dic, keys, leaf_only=False)
        if len(keys) > 1:
            del_dict_empty_item(dic, keys[:-1])








import os
from dsocli.logger import Logger
from dsocli.configs import Config, ContextMode
from dsocli.providers import Providers
from dsocli.templates import TemplateProvider
from dsocli.stages import Stages
from dsocli.constants import *
from dsocli.exceptions import DSOException
from dsocli.local_utils import *
from dsocli.settings import *


__default_spec = {
    'path': os.path.join(Config.config_dir, 'template/'),
}


def get_default_spec():
    return __default_spec.copy()


class LocalTemplateProvider(TemplateProvider):


    def __init__(self):
        super().__init__('template/local/v1')


    @property
    def root_dir(self):
        return Config.template_spec('path')


    def get_path_prefix(self):
        # return self.root_dir + os.sep
        return self.root_dir


    def list(self, uninherited=False, include_contents=False, filter=None):
        Logger.debug(f"Listing local templates: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        templates = load_context_templates(path_prefix=self.get_path_prefix(), uninherited=uninherited, include_contents=include_contents, filter=filter)
        result = []
        for key, details in templates.items():
            item = {'Key': key}
            item.update(details)
            result.append(item)
        return result


    def add(self, key, contents, render_path=None):
        Logger.debug(f"Adding local template '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        response = add_local_template(key=key, path_prefix=self.get_path_prefix(), contents=contents)
        result = {
                'Key': key,
                'Stage': Stages.shorten(Config.stage),
                'Context': response['Context'], 
                'Path': response['Path'],
            }
        return result


    def get(self, key, uninherited=False, revision=None):
        if revision:
            Logger.warn(f"Template provider 'local/v1' does not support versioning. Revision request ignored.")
        Logger.debug(f"Getting template '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        found = locate_template_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(), uninherited=uninherited, include_contents=True)
        if not found:
            raise DSOException(f"Template '{key}' not found in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        result = {
                'Key': key, 
            }
        result.update(found[key])
        return result


    def history(self, key, include_contents=False):
        Logger.warn(f"Template provider 'local/v1' does not support history.")
        Logger.debug(f"Getting template '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        found = locate_template_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(), include_contents=include_contents, uninherited=True)
        if not found:
            raise DSOException(f"Template '{key}' not found in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")

        result = { "Revisions":
            [{
                'RevisionId': '0',
            }]
        }
        result['Revisions'][0].update(found[key])
        return result


    def delete(self, key):
        Logger.debug(f"Locating template: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        ### only parameters owned by the context can be deleted, hence uninherited=True
        found = locate_template_in_context_hierachy(key=key, path_prefix=self.get_path_prefix(), uninherited=True)
        if not found:
            raise DSOException(f"Template '{key}' not found in the given context: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        Logger.info(f"Deleting template: path={found[key]['Path']}")
        delete_local_template(path=f"{found[key]['Path']}/{key}")
        result = {
                'Key': key,
                'Stage': Config.short_stage,
                'Scope': found[key]['Scope'], 
                'Context': found[key]['Context'], 
                'Path': found[key]['Path'],
            }
        return result


def register():
    Providers.register(LocalTemplateProvider())




import os
from .constants import *
from .providers import Providers, ProviderBase
from .logger import Logger
from .configs import Config



class ArtifactoryService():

    def list(self, service, filter=None):
        provider = Providers.ArtifactoryProvider()
        Logger.info(f"Listing artifacts: namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        response = provider.list(filter=filter, service=service)
        result = {
            'Artifacts': response['Files']
        }
        return result

    def add(self, filepath, service, key=None):
        if not key: key = os.path.basename(filepath)
        # self.validate_key(key)
        provider = Providers.ArtifactoryProvider()
        Logger.info(f"Adding artifact '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        return provider.add(filepath=filepath, key=key, service=service)


    def get(self, key, service):
        provider = Providers.ArtifactoryProvider()
        Logger.info(f"Getting artifact '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        return provider.get(key=key, service=service)


    # def history(self, key):
    #     self.AppConfig = AppConfig
    #     provider = Providers.ArtifactoryProvider()
    #     Logger.info(f"Fetching history of artifact '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
    #     return provider.history(key)


    def delete(self, key, service):
        provider = Providers.ArtifactoryProvider()
        Logger.info(f"Deleting artifact '{key}': namespace={Config.get_namespace(ContextMode.Target)}, application={Config.get_application(ContextMode.Target)}, stage={Config.get_stage(ContextMode.Target)}, scope={Config.scope}")
        return provider.delete(key=key, service=service)



Artifactory = ArtifactoryService()



from enum import Enum


class OrderedEnum(Enum):
    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented
    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented
    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented
    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented

    def __str__(self):
        return self.name.split('.')[-1]


    @classmethod
    def from_str(cls, text, case_sensitive=False):
        if case_sensitive:
            items = [item for item in dir(cls) if not item.startswith('_')]
            if text in items:
                return getattr(cls, text)
            return None
        else:
            for item in dir(cls):
                if item.startswith('_'): continue
                if text.lower() == item.lower(): 
                    return getattr(cls, item)
            return None


from .constants import *
from .logger import Logger



class RemoteConfigService():

    def list(self, filter=None, uninherited=False):
        from .providers import Providers
        provider = Providers.RemoteConfigProvider()
        response = provider.list(filter=filter, uninherited=uninherited)
        return response


    def set(self, key, value):
        # self.validate_key(key)
        from .providers import Providers
        provider = Providers.RemoteConfigProvider()
        return provider.set(key=key, value=value)


    def get(self, key, revision=None, uninherited=False, rendered=True):
        from .providers import Providers
        provider = Providers.RemoteConfigProvider()
        return provider.get(key=key, revision=revision, uninherited=uninherited, rendered=rendered)


    def history(self, key):
        from .providers import Providers
        provider = Providers.RemoteConfigProvider()
        return provider.history(key)


    def unset(self, key):
        from .providers import Providers
        provider = Providers.RemoteConfigProvider()
        return provider.unset(key=key)


RemoteConfig = RemoteConfigService()
